/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Paste and Go.
 *
 * The Initial Developer of the Original Code is Jens Bannmann.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Jens Bannmann <jens.b@web.de>
 *  Oliver Aeberhard <aeberhard@gmx.ch>
 *  Holio Lin <gaghiel@gmail.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
var pgISearchSrv = null; // lazy initial

// clipboard object
const pgClipboard = Components.classes["@mozilla.org/widget/clipboard;1"]
    .createInstance(Components.interfaces.nsIClipboard);
const pgTrans = Components.classes["@mozilla.org/widget/transferable;1"]
    .createInstance(Components.interfaces.nsITransferable);
pgTrans.addDataFlavor("text/unicode");

const pgPrefService = Components.classes["@mozilla.org/preferences-service;1"]
    .getService(Components.interfaces.nsIPrefService)

var PasteAndGoCommandId = 'pg2-go-command';
var PasteAndGoKeyId = 'pg2-go-key';
var PasteAndGoMenuitemId = 'pg2-go-menuitem';

var PasteAndSearchCommandId = 'pg2-search-command';
var PasteAndSearchRollCommandId = 'pg2-roll-command';
var PasteAndSearchKeyId = 'pg2-search-key';
var PasteAndSearchMenuitemId = 'pg2-search-menuitem';

var pgStrings;

window.addEventListener("load", function(e) { pgPrefsObserver.startup(); }, false);
window.addEventListener("unload", function(e) { pgPrefsObserver.shutdown(); }, false); // leak avoid

var pgPrefsObserver = {

    prefs: null,
    pgGoKey: null,
    pgSearchKey: null,
    pgGoAccelKeys: null,
    pgSearchAccelKeys: null,

    pgPlaceAtBottom: null,
    pgOpenNewTab: null,
    pgBackgroundTab: null,
    pgPreferSelectionClipboard: null,
    pgHighPriority: null,
    pgRollSearchEngineEnable: null,
    pgSearchEnginesPopupDelayMS: 0,
    pgRollBeginWithCurrent: null,
    pgSwitchAfterRolled: null,

    pgUrlbar: null,
    pgSearchbar: null,

    pgSearchTextbox: null,

    startup: function() {
        pgStrings = document.getElementById("pasteandgo2-strings");

        // Register to receive notifications of preference changes
        this.prefs = pgPrefService.getBranch("extensions.pasteandgo2.");
        this.prefs.QueryInterface(Components.interfaces.nsIPrefBranch2);
        this.prefs.addObserver("", this, false);

        // this.tickerSymbol = this.prefs.getCharPref("symbol");
        this.pgGoKey = this.prefs.getCharPref("GoHotkey");
        this.pgSearchKey = this.prefs.getCharPref("SearchHotkey");
        this.pgGoAccelKeys = this.prefs.getCharPref("GoAccelKeys");
        this.pgSearchAccelKeys = this.prefs.getCharPref("SearchAccelKeys");

        this.pgPlaceAtBottom = this.prefs.getBoolPref("PlaceAtBottom");
        this.pgOpenNewTab = this.prefs.getBoolPref("OpenNewTab");
        this.pgBackgroundTab = this.prefs.getBoolPref("BackgroundTab");
        this.pgPreferSelectionClipboard = this.prefs.getBoolPref("PreferSelectionClipboard");
        this.pgHighPriority = this.prefs.getBoolPref("HighPriority");
        this.pgRollSearchEngineEnable = this.prefs.getBoolPref("RollSearchEngineEnable");
        this.pgSearchEnginesPopupDelayMS = this.prefs.getIntPref("SearchEnginesPopupDelayMS");
        this.pgRollBeginWithCurrent = this.prefs.getBoolPref("RollBeginWithCurrent");
        this.pgSwitchAfterRolled = this.prefs.getBoolPref("SwitchAfterRolled");

        this.pgUrlbar = document.getElementById("urlbar");
        this.pgSearchbar = document.getElementById("searchbar"); // before 1.0, searchbar is 'search-bar'

        if (this.pgUrlbar) {
            dockElement("keyset", "key", PasteAndGoKeyId,
                    new Array("modifiers",   "key",   "command"  ),
                    new Array(this.pgGoAccelKeys, this.pgGoKey, PasteAndGoCommandId));
            dockElement("commandset", "command", PasteAndGoCommandId,
                    new Array("oncommand"),
                    new Array("PasteAndGo();"));
            // have to install on evry time pop up
            this.pgUrlbar.addEventListener("popupshowing", pgInstallGo, false);
        }

        if (this.pgSearchbar) {
            // dock by menuroll
            dockElement("keyset", "key", PasteAndSearchKeyId,
                    new Array("modifiers",   "key",       "command"  ),
                    new Array(this.pgSearchAccelKeys, this.pgSearchKey, PasteAndSearchCommandId));
            dockElement("commandset", "command", PasteAndSearchCommandId, // normal use
                    new Array("oncommand"),
                    new Array("PasteAndSearch();"));

            var contextMenu;
            this.pgSearchTextbox = document.getAnonymousElementByAttribute(this.pgSearchbar,
                    "class", "searchbar-textbox"); // class:"searchbar-textbox"; intersection of 2.0+ and 1.5
            if (! ('mContextMenu' in this.pgSearchTextbox)) { // 2.0+
                var inputbox = document.getAnonymousElementByAttribute(this.pgSearchTextbox,
                        "anonid", "textbox-input-box");
                contextMenu = document.getAnonymousElementByAttribute(inputbox,
                        "anonid", "input-box-contextmenu");
            } else { // 1.5
                contextMenu = this.pgSearchTextbox.mContextMenu;
            }
            // have to install on evry time pop up
            contextMenu.addEventListener("popupshowing", pgInstallSearch, false);

            // install roller
            if (this.pgRollSearchEngineEnable) {
                MenuRoll.startup(this);
                // bug: enable of paste and search by middle click shouldn't ref to pgRollSearchEngineEnable
                MenuRoll.popup.addEventListener('mouseup', onceMouseUpOnPopup, false);
            }

            /*var engineButton =
            document.getAnonymousElementByAttribute(this.pgSearchbar,
                    'anonid', 'searchbar-engine-button'); // 2.0+ works
            engineButton.addEventListener('mouseup',
                function(event) {
                    if (event.button == 1) // middle click
                        PasteAndSearch();
                },
            false);*/
        }

        //this.refreshInformation();
        dockElement("commandset", "command", 'pg2-reload-command',
                new Array("oncommand"),
                new Array("pgPrefsObserver.reload();"));
    },

    shutdown: function() {
        if (this.prefs)
            this.prefs.removeObserver("", this);
        if (this.pgRollSearchEngineEnable)
            MenuRoll.shutdown();
    },

    reload: function() {
        this.shutdown();
        this.startup();
    },

    observe: function(subject, topic, data) {
        if (topic != "nsPref:changed") {
            return;
        }

        // alert('observe: '+ data);
        switch(data) {
            case "GoHotkey":
                this.pgGoKey = this.prefs.getCharPref("GoHotkey");
            break;

            case "SearchHotkey":
                this.pgSearchKey = this.prefs.getCharPref("SearchHotkey");
            break;

            case "GoAccelKeys":
                this.pgGoAccelKeys = this.prefs.getCharPref("GoAccelKeys");
            break;

            case "SearchAccelKeys":
                this.pgSearchAccelKeys = this.prefs.getCharPref("SearchAccelKeys");
            break;

            case "PlaceAtBottom":
                this.pgPlaceAtBottom = this.prefs.getBoolPref("PlaceAtBottom");
            break;

            case "OpenNewTab":
                this.pgOpenNewTab = this.prefs.getBoolPref("OpenNewTab");
            break;

            case "BackgroundTab":
                this.pgBackgroundTab = this.prefs.getBoolPref("BackgroundTab");
            break;

            case "PreferSelectionClipboard":
                this.pgPreferSelectionClipboard = this.prefs.getBoolPref("PreferSelectionClipboard");
            break;

            case "HighPriority":
                this.pgHighPriority = this.prefs.getBoolPref("HighPriority");
            break;

            case "RollSearchEngineEnable":
                this.pgRollSearchEngineEnable = this.prefs.getBoolPref("RollSearchEngineEnable");
            break;

            case "SearchEnginesPopupDelayMS":
                this.pgSearchEnginesPopupDelayMS = this.prefs.getIntPref("SearchEnginesPopupDelayMS");
            break;

            case "RollBeginWithCurrent":
                this.pgRollBeginWithCurrent = this.prefs.getBoolPref("RollBeginWithCurrent");
            break;


            case "SwitchAfterRolled":
                this.pgSwitchAfterRolled = this.prefs.getBoolPref("SwitchAfterRolled");
            break;
        }

        this.refreshInformation();
    },

    refreshInformation: function() {
        //adjustOrderById(PasteAndGoCommandId); // noneed
        //adjustOrderById(PasteAndSearchCommandId); // noneed
        adjustKeyById(PasteAndGoKeyId, this.pgGoAccelKeys, this.pgGoHotKey);
        adjustKeyById(PasteAndSearchKeyId, this.pgSearchAccelKeys, this.pgSearchHotKey);

        // auto re-install when next time menupopup
        removeElementById(PasteAndGoMenuitemId);
        removeElementById(PasteAndSearchMenuitemId);
        removeElementById(PasteAndGoMenuitemId + '-separator');
        removeElementById(PasteAndSearchMenuitemId + '-separator');

        MenuRoll.shutdown();
        if (this.pgRollSearchEngineEnable)
            MenuRoll.startup(this);
    },
};

function adjustKeyById(id, modifiers, key) {
    var element = document.getElementById(id);

    // bug: reassign have to restart
    element.setAttribute("modifiers", modifiers);
    element.setAttribute("key", key);
    adjustOrderByElement(element, pgPrefsObserver.pgHighPriority);
}

function adjustOrderByElement(element, isTop) {
    var parent = element.parentNode;
    parent.removeChild(element);

    // Set position in the keyset table. (If two commands share the
    // same hotkey configuration, only the topmost in the keyset table
    // will be executed)
    if (isTop) {
        parent.insertBefore(element, parent.firstChild);
    } else {
        parent.appendChild(element);
    }
}

function removeElementById(id) {
    var dying = document.getElementById(id);
    if (dying && dying.parentNode)
        dying.parentNode.removeChild(dying);
}

function pgInstallGo(e) {
    pgInstallMenuItem(e, PasteAndGoMenuitemId, pgStrings.getString("extensions.pasteandgo2.goLabel"),
            PasteAndGoCommandId, pgStrings.getString("extensions.pasteandgo2.goAccesskey"));
}

function pgInstallSearch(e) {
    pgInstallMenuItem(e, PasteAndSearchMenuitemId, pgStrings.getString("extensions.pasteandgo2.searchLabel"),
            PasteAndSearchCommandId, pgStrings.getString("extensions.pasteandgo2.searchAccesskey"));
}

function pgInstallMenuItem(e, aID, aLabel, aCommand, aAccesskey) {
    if (e.originalTarget.localName == "menupopup"     // 1. event trigged by menupopup
            && !document.getElementById(aID)) {       // 2. avoid re-install
        var mi = document.createElement("menuitem");
        mi.setAttribute("id", aID);
        mi.setAttribute("label", aLabel);
        mi.setAttribute("command", aCommand);
        mi.setAttribute("accesskey", aAccesskey);
        mi.setAttribute("image", "chrome://pasteandgo2/skin/menuitem-icon.png");
        mi.setAttribute("class", "menuitem-iconic");
        if (pgPrefsObserver.pgPlaceAtBottom) { // place at bottom
            var separator = document.createElement("menuseparator");
            separator.id = aID + '-separator';

            e.originalTarget.appendChild(separator);
            e.originalTarget.appendChild(mi);
        } else {               // else place next to paste
            var items = e.originalTarget.childNodes;
            for (var i=0; i < items.length; i++) {
                var cmd = items[i].getAttribute("cmd");
                if(cmd == "cmd_paste") { // find out paste
                    e.originalTarget.insertBefore(mi, items[i+1]);
                    break;
                }
            }
        }

        return mi;
    }

    return null;
}

// for assign high priority hot key
// create type whos id, pref as parameter said, and dock it under first setType in document
function dockElement(setType, type, id, aNames, aValues) {
    var installed = document.getElementById(id);
    if (installed)
        return;

    var sets = document.getElementsByTagName(setType); // would `by id' be better?
    var set;
    if (sets && sets.length > 0) {
        set = sets[0];
    } else {
        set = document.createElement(setType);
        document.documentElement.appendChild(set);
    }
    /*create a t with type 'type', let t.id = 'id'
      for each aNames, say a,
      each aValues, say v:
      t.a = v*/
    var element = document.createElement(type);
    element.setAttribute("id", id);
    for (i=0; i < aNames.length; i++) {
        element.setAttribute(aNames[i], aValues[i]);
    }
    set.appendChild(element);

    adjustOrderByElement(element, pgPrefsObserver.pgHighPriority);
    return element;
}

function pgGetClipboard(which) {
    pgClipboard.getData(pgTrans, which);

    var str = new Object();
    var strLength = new Object();

    try {
        pgTrans.getTransferData("text/unicode", str, strLength);
    } catch (e) {
        str = null;
    }

    if (str) {
        if ("nsISupportsWString" in Components.interfaces) {
            str = str.value.QueryInterface(Components.interfaces.nsISupportsWString);
        } else if ("nsISupportsString" in Components.interfaces) {
            str = str.value.QueryInterface(Components.interfaces.nsISupportsString);
        } else {
            str = null;
        }
    }

    return str;
}

function getClipboardText() {
    var globalClipboard = pgGetClipboard(pgClipboard.kGlobalClipboard);
    var str = globalClipboard;

    // On Linux, use either selection or global clipboard according to pref and situation
    if (pgClipboard.supportsSelectionClipboard()) {
        var selectionClipboard = pgGetClipboard(pgClipboard.kSelectionClipboard);

        if (selectionClipboard && (pgPrefsObserver.pgPreferSelectionClipboard || !globalClipboard)) {
            str = selectionClipboard;
        }
    }

    return str;
}

function PasteAndGo() {
    var str = getClipboardText();
    pgLoadTab(str);
}

function PasteAndSearch() {
    var cengine = pgPrefsObserver.pgSearchbar.currentEngine;
    if (cengine) // 2.0+
        searchByEngine(cengine);
    else // 1.5
        searchByEngine(pgPrefsObserver.pgSearchTextbox.currentEngine);
}

function searchByEngine(engine) {

    var str = getClipboardText();
    // put the query in the search bar
    pgPrefsObserver.pgSearchTextbox.value = str;

    // add search text as into search history
    var isFormFillEnable = pgPrefService.getBoolPref('browser.formfill.enable');
    if (str && isFormFillEnable) {
        var textBox = pgPrefsObserver.pgSearchTextbox;
        var historyService;
        if ('_formHistSvc' in textBox) // 2.0+
            historyService = textBox._formHistSvc;
        else
            historyService = Components.classes["@mozilla.org/satchel/form-history;1"]
                .getService(Components.interfaces.nsIFormHistory);

        historyService.addEntry(textBox.getAttribute("autocompletesearchparam"), str);
    }

    var postData = null;
    var url = null;
    if ((typeof engine) != "string") { // 2.0+
        // null parameter below specifies HTML response for search
        var submission = engine.getSubmission(pgPrefsObserver.pgSearchbar.value, null);

        if (submission) {
            url = submission.uri.spec;
            postData = submission.postData;
        }

    } else { // 1.5
        if (!pgISearchSrv) // lazy initial
            pgISearchSrv = Components.classes["@mozilla.org/rdf/datasource;1?name=internetsearch"]
                 .getService(Components.interfaces.nsIInternetSearchService);

        url = pgISearchSrv.GetInternetSearchURL(engine,
                encodeURIComponent(str), 0, 0, {value:0});
    }

    pgLoadTab(url, postData);
}

function onceMouseUpOnPopup(event) { // MiddleClickAndSearch
    if (event.button == 1) { // middle click
        var childs = MenuRoll.items; // popup.childNodes;
        var len = childs.length;
        var i;
        for (i = 0; i < len; i++)
            if (childs[i] == event.target)
                break;

        var engines = MenuRoll.getEngines(); //pgPrefsObserver.pgSearchbar._engines;
        if (i < engines.length)
            searchByEngine(engines[i]);

        MenuRoll.popup.hidePopup();
    }
}

function pgLoadTab(aURL, aPostData) {

    var browser = getBrowser();
    var selectedTab = browser.selectedTab;

    // var selectedIsBlank = window._content.document.location.href
    //                        == "about:blank"; // old school
    var selectedIsBlank = (browser.currentURI.spec == 'about:blank');
    var selectedIsUnused = selectedIsBlank && !selectedTab.hasAttribute("busy");
    var shouldOpenNewTab = (pgPrefsObserver.pgOpenNewTab && !selectedIsUnused);

    if (shouldOpenNewTab) {
        var newtab = browser.addTab(aURL, null, null, aPostData, null, true);
        if (! pgPrefsObserver.pgBackgroundTab)
            browser.selectedTab = newtab;
    } else {
        loadURI(aURL, null, aPostData, true);
    }

    content.focus();
}
